// Receives and displays data message formats for RF sensors
// Written by Boomer48: September 2019

// Input port for RF receiver data
#define Pulse_In 2 // INT0

// All times in microseconds
#define Min_Gap_Time 3000
#define Max_Gap_Time 16000
#define Min_Bit_Time 100
#define Max_Bit_Time 1500
#define Uncertainty 200 // Data bit pulse width uncertainty

byte RF_Bit_Count = 0;
byte RF_Byte_Count = 0;
byte RF_Byte = 0;
char ASCII[] = "0123456789ABCDEF";
byte RF_Message[] = {0, 0, 0, 0, 0, 0, 0};
unsigned long Start_Time = 0;
int Pulse_Width;
byte Started = false;
int Gap_Time;
int Bit_Time1; // elapsed time of first bit
int Bit_Time2; // elapsed time of bit that mismatches first bit
byte Invert_Flg; // invert message byte if first bit was actually "0" instead of "1"
byte Msg_Complete;

void setup() {
  Serial.begin(115200);
  digitalWrite(Pulse_In, INPUT_PULLUP); //turn on pullup
  Msg_Complete = false;

  // Enable INT0 external interrupt, trigger on both edges
  bitSet(EICRA, ISC00);
  bitSet(EIMSK, INT0);
}

void loop() {
  byte i;
  byte temp;

  if ((Msg_Complete == true) && (RF_Byte_Count > 2) && (RF_Byte_Count < 8)) {
    noInterrupts();
    Msg_Complete = false;
    Serial.write("Gap_Time: ");
    Serial.print(Gap_Time);
    Serial.write("   Bit_Time1: ");
    Serial.print(Bit_Time1);
    Serial.write("   Bit_Time2 ");
    Serial.print(Bit_Time2);
    Serial.write("   Raw Bytes: ");

    i = 0;
    while (i < RF_Byte_Count) {
      if (Invert_Flg == true)
        RF_Message[i] = ~RF_Message[i];

      Serial.write(ASCII[RF_Message[i] >> 4]); // display upper nibble first
      Serial.write(ASCII[RF_Message[i] & 0x0F]);
      i++;
    }

    Serial.println();
    RF_Byte_Count = 0;
    Started = false;
    PCIFR = 0; // clear all pin change interrupt flags
    interrupts();
  }
} // end loop

// INT0 interrupt handler
ISR (INT0_vect)
{
  //when the pin goes LOW record the pulse start time
  if (digitalRead(Pulse_In) == LOW) {
    Start_Time = micros();
  }
  else // pin went high
    if (Start_Time != 0) {
      //calculate the pulse width
      Pulse_Width = ((volatile int)micros() - Start_Time);
      //clear the timer
      Start_Time = 0;

      // Now check for valid message
      if (Pulse_Width < Max_Gap_Time) {
        if (Started == true) {

          if ((Pulse_Width > Min_Bit_Time) && (Pulse_Width < Max_Bit_Time)) {
            // bit value in range

            if (Bit_Time1 == 0) {
              // first bit
              Bit_Time1 = Pulse_Width;
            }

            if ((Pulse_Width > (Bit_Time1 - Uncertainty)) && (Pulse_Width < (Bit_Time1 + Uncertainty))) {
              // bit = 1
              RF_Bit_Count--; // bits received MSB first
              bitSet(RF_Message[RF_Byte_Count], RF_Bit_Count);
              if (RF_Bit_Count == 0) {
                RF_Byte_Count++;
                RF_Bit_Count = 8;
              }
            }
            else {
              // bit = 0
              RF_Bit_Count--; // bits received MSB first
              bitClear(RF_Message[RF_Byte_Count], RF_Bit_Count);
              if (RF_Bit_Count == 0) {
                RF_Byte_Count++;
                RF_Bit_Count = 8;
              }
              if (Bit_Time2 == 0) {
                // Mismatch of first bit
                Bit_Time2 = Pulse_Width;
                if (Bit_Time1 > Bit_Time2)
                  Invert_Flg = true;
                else
                  Invert_Flg = false;
              }
            }
          }

          else { // bad bit or end of message
            if (Pulse_Width > Min_Gap_Time) {
              Msg_Complete = true;
            }
            Started = false;
          }
        } // Started

        else if (Pulse_Width > Min_Gap_Time) {
          // start of message
          Started = true;
          RF_Byte_Count = 0;
          RF_Bit_Count = 8;
          Gap_Time = Pulse_Width; // save for display
          Bit_Time1 = 0;
          Bit_Time2 = 0;
        }
      }
      else { // invalid message
        Started = false;
      }
    }
}
